import ast
import sys


class GenerationFunctionsAnalyser(ast.NodeVisitor):
    """Receives a codegen script and stores which of the imports come from pystencils_walberla or lbmpy_walberla."""

    def __init__(self):
        self.imports = set()

    def visit_ImportFrom(self, node):
        if "pystencils_walberla" in node.module or "lbmpy_walberla" in node.module:
            for alias in node.names:
                self.imports.add(alias.name)
        self.generic_visit(node)


class CodegenScriptAnalyser(ast.NodeVisitor):
    """Receives a codegen script and stores the names of the output-files that will be generated by the script.

    Args:
        codegen_functions: A list of functions that will be used by the codegen script to generate the files. This
        list can be obtained automatically with the `GenerationFunctionsAnalyser`
    """
    def __init__(self, codegen_functions, gpu=False):
        self.codegen_functions = codegen_functions
        self.functions = list()
        self.gpu = gpu

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name):
            if node.func.id in self.codegen_functions:
                name = node.func.id
                class_name = node.args[1].value
                target = None
                for keyword in node.keywords:
                    if keyword.arg == 'target':
                        target = keyword.value.attr
                for arg in node.args:
                    if hasattr(arg, "value"):
                        if hasattr(arg.value, "attr") and arg.value.attr == "Target":
                            target = arg.attr

                fct = (name, class_name, target)
                if self.gpu:
                    if (name, class_name, "CPU") in self.functions:
                        idx = self.functions.index((name, class_name, "CPU"))
                        self.functions[idx] = (name, class_name, target)
                    else:
                        self.functions.append(fct)
                else:
                    if target != "GPU":
                        self.functions.append(fct)
        self.generic_visit(node)


r = open('/Users/holzer/walberla/apps/benchmarks/UniformGridCPU/UniformGridCPU.py', 'r')
t = ast.parse(r.read())


import_analyzer = GenerationFunctionsAnalyser()
import_analyzer.visit(t)
codegen_functions = import_analyzer.imports

if 'CodeGeneration' in codegen_functions:
    codegen_functions.remove('CodeGeneration')

analyzer = CodegenScriptAnalyser(codegen_functions)
analyzer.visit(t)

file_list = []
for fct in analyzer.functions:
    if fct[1] == "UniformGridCPU_PackInfo":
        file_list.append(fct[1] + "Even.cpp")
        file_list.append(fct[1] + "Even.h")
        file_list.append(fct[1] + "Odd.cpp")
        file_list.append(fct[1] + "Odd.h")
    else:
        if fct[2] == "CPU":
            file_list.append(fct[1] + ".cpp")
            file_list.append(fct[1] + ".h")
        elif fct[2] == "GPU":
            file_list.append(fct[1] + ".h")
            file_list.append(fct[1] + ".cu")
        elif fct[2] is None:
            file_list.append(fct[1] + ".h")

print(";".join(file_list))
